
module Development.Shake.Internal.History.Types(
    BS_QTypeRep, BS_Key, BS_Store, BS_Identity
    ) where

import qualified Data.ByteString as BS

type BS_QTypeRep = BS.ByteString
type BS_Key = BS.ByteString
type BS_Store = BS.ByteString
type BS_Identity = BS.ByteString
